import inspect
import warnings
import pysb
from pysb.core import MultiState
import sympy
from sympy.printing import StrPrinter
from sympy.printing.precedence import precedence


class BngGenerator(object):
    def __init__(self, model, additional_initials=None, population_maps=None):
        self.model = model
        if additional_initials is None:
            additional_initials = []
        self._additional_initials = additional_initials
        self._population_maps = population_maps
        self.__content = None

    def get_content(self):
        if self.__content == None:
            self.generate_content()
        return self.__content

    def generate_content(self):
        self.__content = "begin model\n"
        self.generate_parameters()
        self.generate_compartments()
        self.generate_molecule_types()
        self.generate_observables()
        self.generate_functions()
        self.generate_species()
        self.generate_reaction_rules()
        self.generate_population_maps()
        self.__content += "end model\n"

    def generate_parameters(self):
        exprs = self.model.expressions_constant()
        if not self.model.parameters and not exprs:
            return
        self.__content += "begin parameters\n"
        max_length = max(len(p.name) for p in
                         self.model.parameters | self.model.expressions)
        for p in self.model.parameters:
            self.__content += (("  %-" + str(max_length) + "s   %.17g\n") %
                               (p.name, p.value))
        for e in exprs:
            self.__content += (("  %-" + str(max_length) + "s   %s\n") %
                               (e.name, expression_to_muparser(e)))
        self.__content += "end parameters\n\n"

    def generate_compartments(self):
        if not self.model.compartments:
            return
        self.__content += "begin compartments\n"
        for c in self.model.compartments:
            if c.parent is None:
                parent_name = ''
            else:
                parent_name = c.parent.name
            if c.size is None:
                size = "1.0"
            else:
                size = c.size.name
            self.__content += ("  %s  %d  %s  %s\n") % (c.name, c.dimension, size, parent_name)
        self.__content += "end compartments\n\n"        

    def generate_molecule_types(self):
        if not self.model.monomers:
            return
        self.__content += "begin molecule types\n"
        for m in self.model.monomers:
            site_code = ','.join([format_monomer_site(m, s) for s in m.sites])
            self.__content += "  %s(%s)\n" % (m.name, site_code)
        self.__content += "end molecule types\n\n"

    def generate_reaction_rules(self):
        if not self.model.rules:
            warn_caller("Model does not contain any rules")
            return
        self.__content += "begin reaction rules\n"
        max_length = max(len(r.name) for r in self.model.rules) + 1  # +1 for the colon
        for r in self.model.rules:
            label = r.name + ':'
            react_p = r.reactant_pattern
            prod_p = r.product_pattern
            if not react_p.complex_patterns:
                react_p = None
            if not prod_p.complex_patterns:
                prod_p = None
            reactants_code = format_reactionpattern(react_p)
            products_code = format_reactionpattern(prod_p)
            arrow = '->'
            if r.is_reversible:
                arrow = '<->'
            self.__content += ("  %-" + str(max_length) + "s  %s %s %s    %s") % \
                (label, reactants_code, arrow, products_code, r.rate_forward.name)
            self.__content += _tags_in_rate(r.rate_forward)
            if r.is_reversible:
                self.__content += ', %s' % r.rate_reverse.name
                self.__content += _tags_in_rate(r.rate_reverse)
            if r.delete_molecules:
                self.__content += ' DeleteMolecules'
            if r.move_connected:
                self.__content += ' MoveConnected'
            self.__content += "\n"
        self.__content += "end reaction rules\n\n"

    def generate_observables(self):
        if not self.model.observables:
            return
        max_length = max(len(name) for name in self.model.observables.keys())
        self.__content += "begin observables\n"
        for obs in self.model.observables:
            observable_code = format_reactionpattern(obs.reaction_pattern,
                                                     for_observable=True)
            self.__content += ("  %s %-" + str(max_length) + "s   %s\n") % \
                              (obs.match.title(), obs.name, observable_code)
        self.__content += "end observables\n\n"

    def generate_functions(self):
        exprs = self.model.expressions_dynamic()
        if not exprs:
            return
        max_length = max(len(e.name) for e in exprs) + 2
        self.__content += "begin functions\n"
        for i, e in enumerate(exprs):
            signature = '{}({})'.format(e.name, ','.join(sorted([sym.name for sym in e.expr.atoms(pysb.Tag)])))
            self.__content += ("  %-" + str(max_length) + "s   %s\n") % \
                (signature, expression_to_muparser(e))
        self.__content += "end functions\n\n"

    def generate_species(self):
        if not self.model.initials:
            warn_caller("Model does not contain any initial conditions")
            return
        species_codes = [
            format_complexpattern(ic.pattern, ic.fixed)
            for ic in self.model.initials
        ]
        for cp in self._additional_initials:
            if not any([
                cp.is_equivalent_to(ic.pattern) for ic in self.model.initials
            ]):
                species_codes.append(format_complexpattern(cp))
        max_length = max(len(code) for code in species_codes)
        self.__content += "begin species\n"
        for i, code in enumerate(species_codes):
            if i < len(self.model.initials):
                param = self.model.initials[i].value.name
            else:
                param = '0'
            self.__content += ("  %-" + str(max_length) + "s   %s\n") % (code,
                                                                         param)
        self.__content += "end species\n\n"

    def generate_population_maps(self):
        if self._population_maps is None:
            return
        self.__content += 'begin population maps\n'
        cplx_pats = [format_complexpattern(pm.complex_pattern)
                     for pm in self._population_maps]
        str_padding = max(len(cp) for cp in cplx_pats)
        for i in range(len(cplx_pats)):
            cs = self._population_maps[i].counter_species
            if cs is None:
                cs = '__hppcounter%d()' % i
                self._population_maps[i].counter_species = cs
            self.__content += ('  %-' + str(str_padding) + 's -> %s\t%s\n') % \
                              (cplx_pats[i], cs,
                               self._population_maps[i].lumping_rate.name)
        self.__content += 'end population maps\n\n'


def _tags_in_rate(expr):
    if not isinstance(expr, pysb.Expression):
        return ''

    tags = expr.tags()

    return '({})'.format(', '.join([t.name for t in tags]))


def format_monomer_site(monomer, site):
    ret = site
    if site in monomer.site_states:
        for state in monomer.site_states[site]:
            ret += '~' + state
    return ret

def format_reactionpattern(rp, for_observable=False):
    if rp is None:
        return '0'
    if for_observable is False:
        delimiter = ' + '
    else:
        delimiter = ' '
    return delimiter.join([format_complexpattern(cp) for cp in rp.complex_patterns])

def format_complexpattern(cp, fixed=False):
    if cp is None:
        return '0'
    ret = '.'.join([format_monomerpattern(mp) for mp in cp.monomer_patterns])
    if fixed:
        ret = '$' + ret
    if cp.compartment is not None:
        ret = '@%s:%s' % (cp.compartment.name, ret)
    if cp._tag:
        ret = '%{}:{}'.format(cp._tag.name, ret)
    if cp.match_once:
        ret = '{MatchOnce}' + ret
    return ret

def format_monomerpattern(mp):
    # sort sites in the same order given in the original Monomer
    site_conditions = sorted(mp.site_conditions.items(),
                             key=lambda x: mp.monomer.sites.index(x[0]))
    site_pattern_code = ','.join([format_site_condition(site, state) for (site, state) in site_conditions])
    ret = '%s(%s)' % (mp.monomer.name, site_pattern_code)
    if mp.compartment is not None:
        ret = '%s@%s' % (ret, mp.compartment.name)
    if mp._tag:
        ret = '{}%{}'.format(ret, mp._tag.name)
    return ret

def format_site_condition(site, state):
    # empty
    if state == None:
        state_code = ''
    # single bond
    elif isinstance(state, int):
        state_code = '!' + str(state)
    # multiple bonds
    elif isinstance(state, list) and all(isinstance(s, int) for s in state):
        state_code = ''.join('!%d' % s for s in state)
    # state
    elif isinstance(state, str):
        state_code = '~' + state
    # state AND single bond
    elif isinstance(state, tuple):
        # bond is wildcard (zero or more unspecified bonds)
        if state[1] == pysb.WILD:
            state = (state[0], '?')
        elif state[1] == pysb.ANY:
            state = (state[0], '+')
        state_code = '~%s!%s' % state
    elif isinstance(state, MultiState):
        return ','.join(format_site_condition(site, s) for s in state)
    # one or more unspecified bonds
    elif state is pysb.ANY:
        state_code = '!+'
    # anything at all (usually you can leverage don't-care-don't-write, but in
    # some cases such as when a rule explicitly sets the state of site A but
    # conditions on site B, site A on the reactant side must be set to WILD)
    elif state is pysb.WILD:
        state_code = '!?'
    else:
        raise ValueError("BNG generator has encountered an unknown element in "
                         "a rule pattern site condition.")
    return '%s%s' % (site, state_code)

def warn_caller(message):
    caller_frame = inspect.currentframe()
    # walk up through the stack until we are outside of pysb
    stacklevel = 1
    module = inspect.getmodule(caller_frame)
    while module and module.__name__.startswith('pysb.'):
        stacklevel += 1
        caller_frame = caller_frame.f_back
        module = inspect.getmodule(caller_frame)
    warnings.warn(message, stacklevel=stacklevel)


class BngPrinter(StrPrinter):
    def __init__(self, **settings):
        super(BngPrinter, self).__init__(settings)

    def _print_Piecewise(self, expr):
        if expr.args[-1][1] is not sympy.true:
            raise NotImplementedError('Piecewise statements are only '
                                      'supported if convertible to BNG if '
                                      'statements')

        if_stmt = expr.args[-1][0]
        for pos in range(len(expr.args) - 2, -1, -1):
            if_stmt = 'if({},{},{})'.format(
                self._print(expr.args[pos][1]),
                self._print(expr.args[pos][0]),
                self._print(if_stmt)
            )

        return if_stmt

    def _print_Dummy(self, expr):
        return expr.name

    def _print_Pow(self, expr, rational=False):
        return super(BngPrinter, self)._print_Pow(expr, rational)\
            .replace('**', '^')

    def _print_And(self, expr):
        return super(BngPrinter, self)._print_And(expr).replace('&', '&&')

    def _print_Or(self, expr):
        return super(BngPrinter, self)._print_Or(expr).replace('|', '||')

    def _print_Relational(self, expr):
        if getattr(expr, "rel_op", None) not in {"==", "!=", "<", "<=", ">", ">="}:
            raise NotImplementedError(
                "Relational operator not supported: %s" % type(expr).__name__
            )
        # Adapted from StrPrinter._print_Relational.
        return '%s %s %s' % (
            self.parenthesize(expr.lhs, precedence(expr)),
            expr.rel_op,
            self.parenthesize(expr.rhs, precedence(expr)),
        )

    def _print_log(self, expr):
        # BNG doesn't accept "log", only "ln".
        return 'ln' + "(%s)" % self.stringify(expr.args, ", ")

    def _print_Pi(self, expr):
        return '_pi'

    def _print_Exp1(self, expr):
        return '_e'

    def _print_floor(self, expr):
        return 'rint({} - 0.5)'.format(self._print(expr.args[0]))

    def _print_ceiling(self, expr):
        return '(rint({} + 1) - 1)'.format(self._print(expr.args[0]))

    def __make_lower(self, expr):
        """ Print a function with its name in lower case """
        return '{}({})'.format(
            self._print(expr.func).lower(),
            self._print(expr.args[0] if len(expr.args) == 1 else
                        ', '.join([self._print(a) for a in expr.args]))
        )

    _print_Abs = __make_lower
    _print_Min = __make_lower
    _print_Max = __make_lower


def expression_to_muparser(expression):
    """Render the Expression as a muparser-compatible string."""

    return BngPrinter(order='none').doprint(expression.expr)



#!/usr/bin/env python
"""
Usage: ``python -m pysb.tools.render_species mymodel.py > mymodel.dot``

Renders the species from a model into the "dot" graph format which can be
visualized with Graphviz.

To create a PDF from the .dot file, use the Graphviz tools in the following
command pipeline::

    ccomps -x mymodel.dot | dot | gvpack -m0 | neato -n2 -T pdf -o mymodel.pdf

You can also change the "dot" command to "circo" or "sfdp" for a different type
of layout. Note that you can pipe the output of render_species straight into a
Graphviz command pipeline without creating an intermediate .dot file, which is
especially helpful if you are making continuous changes to the model and need to
visualize your changes repeatedly::

    python -m pysb.tools.render_species mymodel.py | ccomps -x | dot |
      gvpack -m0 | neato -n2 -T pdf -o mymodel.pdf

Note that some PDF viewers will auto-reload a changed PDF, so you may not even
need to manually reopen it every time you rerun the tool.
"""

import sys
import os
import re
try:
    import pygraphviz
except ImportError:
    pygraphviz = None
import pysb.bng


def run(model):
    """
    Render the species from a model into the "dot" graph format.

    Parameters
    ----------
    model : pysb.core.Model
        The model to render.

    Returns
    -------
    string
        The dot format output.
    """

    pysb.bng.generate_equations(model)
    return render_species_as_dot(model.species, model.name)


def render_species_as_dot(species_list, graph_name=""):
    if pygraphviz is None:
        raise ImportError('pygraphviz library is required to run this '
                          'function')
    graph = pygraphviz.AGraph(name="%s species" % graph_name, rankdir="LR",
                              fontname='Arial')
    graph.edge_attr.update(fontname='Arial', fontsize=8)
    for si, cp in enumerate(species_list):
        sgraph_name = 'cluster_s%d' % si
        cp_label = re.sub(r'% ', '%<br align="left"/>',
                          str(cp)) + '<br align="left"/>'
        sgraph_label = '<<font point-size="10" color="blue">s%d</font>' \
                       '<br align="left"/>' \
                       '<font face="Consolas" point-size="6">%s</font>>' % \
                       (si, cp_label)
        sgraph = graph.add_subgraph(name=sgraph_name, label=sgraph_label,
                                    color="gray75", sortv=sgraph_name)
        bonds = {}
        for mi, mp in enumerate(cp.monomer_patterns):
            monomer_node = '%s_%d' % (sgraph_name, mi)
            monomer_label = '<<table border="0" cellborder="1" cellspacing="0">'
            monomer_label += '<tr><td bgcolor="#a0ffa0"><b>%s</b></td></tr>' % \
                             mp.monomer.name
            for site in mp.monomer.sites:
                site_state = None
                cond = mp.site_conditions[site]
                if isinstance(cond, str):
                    site_state = cond
                elif isinstance(cond, tuple):
                    site_state = cond[0]
                site_label = site
                if site_state is not None:
                    site_label += '=<font color="purple">%s</font>' % site_state
                monomer_label += '<tr><td port="%s">%s</td></tr>' % \
                                 (site, site_label)
            for site, value in mp.site_conditions.items():
                site_bonds = []  # list of bond numbers
                if isinstance(value, int):
                    site_bonds.append(value)
                elif isinstance(value, tuple):
                    site_bonds.append(value[1])
                elif isinstance(value, list):
                    site_bonds += value
                for b in site_bonds:
                    bonds.setdefault(b, []).append((monomer_node, site))
            monomer_label += '</table>>'
            sgraph.add_node(monomer_node,
                            label=monomer_label, shape="none", fontname="Arial",
                            fontsize=8)
        for bi, sites in bonds.items():
            node_names, port_names = list(zip(*sites))
            sgraph.add_edge(node_names, tailport=port_names[0],
                            headport=port_names[1], label=str(bi))
    return graph.string()

usage = __doc__
usage = usage[1:]  # strip leading newline

if __name__ == '__main__':
    # sanity checks on filename
    if len(sys.argv) <= 1:
        print(usage, end=' ')
        exit()
    model_filename = sys.argv[1]
    if not os.path.exists(model_filename):
        raise Exception("File '%s' doesn't exist" % model_filename)
    if not re.search(r'\.py$', model_filename):
        raise Exception("File '%s' is not a .py file" % model_filename)
    sys.path.insert(0, os.path.dirname(model_filename))
    model_name = re.sub(r'\.py$', '', os.path.basename(model_filename))
    # import it
    try:
        # FIXME if the model has the same name as some other "real" module
        # which we use, there will be trouble (use the imp package and import
        # as some safe name?)
        model_module = __import__(model_name)
    except Exception as e:
        print("Error in model script:\n")
        raise
    # grab the 'model' variable from the module
    try:
        model = model_module.__dict__['model']
    except KeyError:
        raise Exception("File '%s' isn't a model file" % model_filename)
    print(run(model))



"""
A wrapper class, ``Builder``, that facilitates the programmatic construction of
PySB models while adding a few features useful for model calibration.

The pattern for model construction using this class does not rely on the
SelfExporter class of PySB. Instead, the ``Builder`` class contains an instance
of a PySB model object. Monomers, Parameters, Rules, etc. are added to this
model object by invoking the wrapper methods of the class, which include

- :py:meth:`pysb.builder.Builder.monomer`
- :py:meth:`pysb.builder.Builder.parameter`
- :py:meth:`pysb.builder.Builder.rule`
- :py:meth:`pysb.builder.Builder.compartment`
- :py:meth:`pysb.builder.Builder.initial`
- :py:meth:`pysb.builder.Builder.observable`

Each of these functions invokes the appropriate PySB component constructor
while setting the argument ``_export=False`` so that the SelfExporter is not
invoked. The created components are then added to the instance of the model
contained in the builder class.

In addition, the model builder class implements the ``__getitem__`` method so
that invoking ``self['component_name']`` from within any method returns the
component with the given name.

In addition to managing model components, the method for parameter creation,
:py:meth:`psyb.builder.Builder.parameter`, allows the caller to specify
whether the parameter should be estimated during model calibration, and if
so, what the prior distribution should be.

Creating custom model builders
------------------------------

A useful application of the Builder class is to create subclasses that
implement macros or motifs for combinatorial model building.  For instance, a
subclass could be created as follows::

    class MyBuilder(pysb.builder.Builder):
        # Constructor with monomer declarations, etc....

        def my_motif():
            k1 = self.parameter('k1', 1, _estimate=False)
            k2 = self.parameter('k2', 1, prior=Uniform(-5, -1))
            A = self['A']
            B = self['B']
            self.rule('A_B_bind', A(b=None) + B(b=None) | A(b=1) % B(b=1),
                      k1, k2)

This motif, implemented as a method, does several things: it manages the
addition of components to the model, keeps track of which parameters to
estimate (``k2`` but not ``k1``), and specifies the prior distribution to use
for ``k2``.

A builder subclass can thus be written containing several motifs, different
subsets of which can be called to build up models in a combinatorial
fashion. Furthermore, implementations of specific motifs can be overridden by
creating additional subclasses.
"""

from pysb import *

class Builder(object):

    # -- CONSTRUCTOR AND MONOMER DECLARATIONS --------------------------------
    def __init__(self, params_dict=None):
        """Base constructor for all model builder classes.

        Initializes collections of the parameters to estimate, as well
        as the means and variances of their priors.

        Parameters
        ----------
        params_dict : dict
            The params_dict allows any parameter value to be overriden
            by name; any parameters not included in the dict will be set
            to default values. For example, if params_dict contains::

                {'tBid_Bax_kf': 1e-2}

            then the parameter tBid_Bax_kf will be assigned a value of 1e-2;
            all other parameters will take on default values. However,
            note that the parameter value given will be multiplied by any
            scaling factor passed in when the parameter is declared.
        """

        self.model = Model('model', _export=False)
        """The PySB model to which components are added."""
        self.estimate_params = []
        """A list of the parameters to be estimated."""
        self.params_dict = params_dict
        """A dict of parameter values to override default values."""
        self.priors = []
        """A list of priors for use in sampling parameter values."""

    # -- CONSTRUCTOR WRAPPER FUNCTIONS ---------------------------------------
    def monomer(self, *args, **kwargs):
        """Adds a parameter to the Builder's model instance."""
        m = Monomer(*args, _export=False, **kwargs)
        self.model.add_component(m)
        return m

    def parameter(self, name, value, factor=1, prior=None):
        """Adds a parameter to the Builder's model instance.

        Examines the params_dict attribute of the Builder instance (which is
        set in the constructor,
        :py:meth:`pysb.builder.Builder.__init__`).  If the
        parameter with the given name is in the ``params_dict``, then the value
        in the ``params_dict`` is used to construct the parameter, and the
        argument ``value`` is ignored. If the parameter is not in
        ``params_dict``, then the parameter is assigned ``value``.

        Furthermore, in all cases the parameter value is multiplied by a
        scaling factor specified by the argument ``factor``. This allows rate
        scaling factor that are dependent on model implementations or on units
        (deterministic vs. stochastic) to be handled by keeping the same
        parameter value but passing in the appropriate value for ``factor``.

        Parameters
        ----------
        name : string
            The name of the parameter to add
        value : number
            The value of the parameter
        factor : number
            A scaling factor to be applied to the parameter value.
        prior : instance of prior class from bayessb.priors
            The prior object describing the prior probability of different
            values for this parameter, if the parameter should be included among
            the parameters to estimate (contained in the set
            ``Builder.estimate_params``).
        """

        if self.params_dict is None:
            param_val = value * factor
        else:
            if name in self.params_dict:
                param_val = self.params_dict[name] * factor
            else:
                param_val = value * factor

        p = Parameter(name, param_val, _export=False)
        self.model.add_component(p)

        if prior is not None:
            self.estimate_params.append(p)
            self.priors.append(prior)

        return p

    def rule(self, *args, **kwargs):
        """Adds a rule to the Builder's model instance."""
        r = Rule(*args, _export=False, **kwargs)
        self.model.add_component(r)
        return r 

    def compartment(self, *args, **kwargs):
        """Adds a compartment to the Builder's model instance."""
        c = Compartment(*args, _export=False, **kwargs)
        self.model.add_component(c)
        return c 

    def observable(self, *args, **kwargs):
        """Adds an observable to the Builder's model instance."""
        o = Observable(*args, _export=False, **kwargs)
        self.model.add_component(o)
        return o

    def expression(self, *args, **kwargs):
        """Adds an expression to the Builder's model instance."""
        e = Expression(*args, _export=False, **kwargs)
        self.model.add_component(e)
        return e

    def initial(self, *args, **kwargs):
        """Adds an initial condition to the Builder's model instance."""
        i = Initial(*args, _export=False, **kwargs)
        self.model.add_initial(i)
        return i

    def tag(self, *args, **kwargs):
        """Adds a tag to the Builder's model instance."""
        t = Tag(*args, _export=False, **kwargs)
        self.model.add_component(t)
        return t

    def __getitem__(self, index):
        """Returns the component with the given string index
        from the instance of the model contained by the Builder."""
        return self.model.all_components()[index]





from json import JSONDecoder
from pysb.builder import Builder
from pysb.core import RuleExpression, ReactionPattern, ComplexPattern, \
    MonomerPattern, MultiState, ANY, WILD, Parameter, Expression
from pysb.annotation import Annotation
from pysb.pattern import SpeciesPatternMatcher
import sympy
import collections
from collections.abc import Mapping
import json
import re
import warnings
from sympy.parsing.sympy_parser import parse_expr


class PySBJSONDecodeError(ValueError):
    pass


class PySBJSONDecoder(JSONDecoder):
    """
    Decode a JSON-encoded PySB model

    See :py:mod:`pysb.export.json` for implementation details.
    """
    MAX_SUPPORTED_PROTOCOL = 1

    def _modelget(self, name):
        if name is None:
            return None
        if name == 'model':
            return self.b.model

        return self.b.model.components[name]

    def decode_state_value(self, sv):
        if isinstance(sv, Mapping) and '__object__' in sv:
            if sv['__object__'] == '__multistate__':
                return MultiState(*sv['sites'])
            if sv['__object__'] == 'ANY':
                return ANY
            if sv['__object__'] == 'WILD':
                return WILD
        try:
            if len(sv) == 2 and isinstance(sv[0], str) \
                    and isinstance(sv[1], (int, Mapping)):
                return sv[0], self.decode_state_value(sv[1])
        except TypeError:
            pass
        return sv

    def decode_monomer(self, mon):
        self.b.monomer(mon['name'], mon['sites'], mon['states'])

    def decode_parameter(self, par):
        self.b.parameter(par['name'], par['value'])

    def decode_derived_parameter(self, par):
        self.b.model._derived_parameters.add(
            Parameter(par['name'], par['value'], _export=False)
        )

    def _parse_expr(self, e, e_name):
        # Quick security check on the expression
        if not re.match(r'^[\w\s()/+\-._*]*$', e):
            raise PySBJSONDecodeError(
                'Security check on %s failed' % e_name
            )

        expr_symbols = {
            s.name: s for s in
            (self.b.model.parameters | self.b.model.expressions |
             self.b.model.tags | self.b.model._derived_parameters |
             self.b.model._derived_expressions)
        }

        expression = parse_expr(e, local_dict=expr_symbols)

        # Replace observables now, to enable expand_expr()
        # Replacing as part of expr_symbols breaks local functions!
        expression = expression.xreplace(
            {sympy.Symbol(s.name): s for s in self.b.model.observables})

        return expression

    def decode_expression(self, expr, derived=False):
        expression = self._parse_expr(expr['expr'],
                                      'expression "{}"'.format(expr['name']))

        if derived:
            self.b.model._derived_expressions.add(
                Expression(expr['name'], expression, _export=False)
            )
        else:
            self.b.expression(
                str(expr['name']),  # Need str() to avoid unicode errors on Py2
                expression
            )

    def decode_derived_expression(self, expr):
        return self.decode_expression(expr, derived=True)

    def decode_observable(self, obs):
        o = self.b.observable(
            obs['name'],
            self.decode_reaction_pattern(obs['reaction_pattern']),
            obs['match']
        )
        try:
            o.coefficients = obs['coefficients']
            o.species = obs['species']
        except KeyError:
            pass

    def decode_monomer_pattern(self, mp):
        mon = self._modelget(mp['monomer'])
        mp_obj = MonomerPattern(
            mon,
            {site: self.decode_state_value(sv)
             for site, sv in mp['site_conditions'].items()},
            self._modelget(mp['compartment'])
        )
        mp_obj._tag = self._modelget(mp['tag'])
        return mp_obj

    def decode_complex_pattern(self, cp):
        cp_obj = ComplexPattern(
            [self.decode_monomer_pattern(mp) for mp in cp['monomer_patterns']],
            self._modelget(cp['compartment']),
            cp['match_once']
        )
        cp_obj._tag = self._modelget(cp['tag'])
        return cp_obj

    def decode_species(self, sp):
        self.b.model.species.append(self.decode_complex_pattern(sp))

    def decode_reaction_pattern(self, rp):
        return ReactionPattern(
            [self.decode_complex_pattern(cp) for cp in rp['complex_patterns']]
        )

    def decode_rule_expression(self, rexp):
        return RuleExpression(
            self.decode_reaction_pattern(rexp['reactant_pattern']),
            self.decode_reaction_pattern(rexp['product_pattern']),
            rexp['reversible']
        )

    def decode_rule(self, r):
        self.b.rule(
            r['name'],
            self.decode_rule_expression(r['rule_expression']),
            self._modelget(r['rate_forward']),
            self._modelget(r['rate_reverse']),
            r['delete_molecules'],
            r['move_connected']
        )

    def decode_tag(self, tag):
        self.b.tag(tag['name'])

    def decode_compartment(self, cpt):
        self.b.compartment(
            cpt['name'],
            self._modelget(cpt['parent']),
            cpt['dimension'],
            self._modelget(cpt['size'])
        )

    def decode_initial(self, init):
        self.b.initial(
            self.decode_complex_pattern(init['pattern']),
            self._modelget(init['parameter_or_expression']),
            init['fixed']
        )

    def decode_annotation(self, ann):
        self.b.model.add_annotation(
            Annotation(
                self._modelget(ann['subject']),
                ann['object'],
                ann['predicate'],
                _export=False
            )
        )

    def decode_reaction(self, rxn):
        rxn['rate'] = self._parse_expr(rxn['rate'], 'reaction rate')
        self.b.model.reactions.append(rxn)

    def decode(self, s):
        res = super(PySBJSONDecoder, self).decode(s)

        if not isinstance(res, dict):
            raise PySBJSONDecodeError('Decode error (not dictionary)')
        if 'protocol' not in res:
            raise PySBJSONDecodeError(
                'No "protocol" entry found - is this a PySB model?')
        if not isinstance(res['protocol'], int):
            raise PySBJSONDecodeError('"protocol" attribute is not an integer')

        if res['protocol'] < 1:
            raise PySBJSONDecodeError('"Invalid "protocol" value: {}'.format(
                res['protocol']))

        if res['protocol'] > self.MAX_SUPPORTED_PROTOCOL:
            raise PySBJSONDecodeError(
                'Protocol {} is not supported (max: {})'.format(
                    res['protocol'], self.MAX_SUPPORTED_PROTOCOL
                )
            )

        self.b = Builder()
        self.b.model.name = res['name']

        decoders = collections.OrderedDict((
            ('monomers', self.decode_monomer),
            ('parameters', self.decode_parameter),
            ('_derived_parameters', self.decode_derived_parameter),
            ('compartments', self.decode_compartment),
            ('observables', self.decode_observable),
            ('tags', self.decode_tag),
            ('expressions', self.decode_expression),
            ('_derived_expressions', self.decode_derived_expression),
            ('rules', self.decode_rule),
            ('initials', self.decode_initial),
            ('annotations', self.decode_annotation),
            ('reactions', self.decode_reaction),
            ('reactions_bidirectional', self.decode_reaction),
            ('species', self.decode_species)
        ))

        for component_type, decoder in decoders.items():
            for component in res.get(component_type, []):
                decoder(component)

        if self.b.model.reactions and self.b.model.observables \
                and 'species' not in res['observables'][0]:

            # We have network, need to regenerate Observable species and coeffs
            warnings.warn(
                'This SimulationResult file is missing Observable species and '
                'coefficients data. These will be generated now - we recommend '
                'you re-save your SimulationResult file to avoid this warning.'
            )

            for obs in self.b.model.observables:
                if obs.match in ('molecules', 'species'):
                    obs_matches = SpeciesPatternMatcher(self.b.model).match(
                        obs.reaction_pattern, index=True, counts=True)
                    sp, vals = zip(*sorted(obs_matches.items()))
                    obs.species = list(sp)
                    if obs.match == 'molecules':
                        obs.coefficients = list(vals)
                    else:
                        obs.coefficients = [1] * len(obs_matches.values())
                else:
                    raise ValueError(f'Unknown obs.match value: {obs.match}')

        return self.b.model


def model_from_json(json_str):
    return json.loads(json_str, cls=PySBJSONDecoder)



from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup
import versioneer
import os


def main():
    this_directory = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(this_directory, 'README.rst'), 'r') as f:
        long_description = f.read()

    cmdclass = versioneer.get_cmdclass()

    setup(name='pysb',
          version=versioneer.get_version(),
          description='Python Systems Biology modeling framework',
          long_description=long_description,
          long_description_content_type='text/x-rst',
          author='Jeremy Muhlich',
          author_email='jmuhlich@bitflood.org',
          url='http://pysb.org/',
          packages=['pysb', 'pysb.generator', 'pysb.importers', 'pysb.tools',
                    'pysb.examples', 'pysb.export', 'pysb.simulator',
                    'pysb.testing', 'pysb.tests'],
          scripts=['scripts/pysb_export'],
          # We should really specify some minimum versions here.
          python_requires='>=3.6',
          install_requires=['numpy', 'scipy>=1.1', 'sympy>=1.6', 'networkx',
                            'futures; python_version == "2.7"'],
          setup_requires=['nose'],
          tests_require=['coverage', 'pygraphviz', 'matplotlib', 'pexpect',
                         'pandas', 'h5py', 'mock', 'cython',
                         'python-libsbml', 'libroadrunner'],
          cmdclass=cmdclass,
          keywords=['systems', 'biology', 'model', 'rules'],
          classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Mathematics',
            ],
          )


if __name__ == '__main__':
    main()



# Based on the BioNetGen model 'localfunc.bngl'
# https://github.com/RuleWorld/bionetgen/blob/master/bng2/Models2/localfunc.bngl
#
# This model demonstrates the use of MultiState and local functions in PySB
#
# Requires Python 3.x or greater (will give a SyntaxError on Python 2.7)

from pysb import Model, Monomer, Parameter, Expression, Rule, \
    Observable, Initial, Tag, MultiState

Model()

Monomer('A', ['b', 'b', 'b'])
Monomer('B', ['a'])
Monomer('C')

Parameter('kp', 0.5)
Parameter('km', 0.1)
Parameter('k_synthC', 1e3)
Parameter('k_degrC', 0.5)
Parameter('Ab_b_b_0', 1.0)
Parameter('Ba_0', 3.0)
Parameter('C_0', 0.0)

Observable('Atot', A())
Observable('Btot', B())
Observable('Ctot', C())
Observable('AB0', A(b=MultiState(None, None, None)), match='species')
Observable('AB1', A(b=MultiState(1, None, None)) % B(a=1), match='species')
Observable('AB2', A(b=MultiState(1, 2, None)) % B(a=1) % B(a=2),
           match='species')
Observable('AB3', A(b=MultiState(1, 2, 3)) % B(a=1) % B(a=2) % B(a=3),
           match='species')
Observable('AB_motif', A(b=1) % B(a=1))

Tag('x')

Expression('f_synth', k_synthC * AB_motif(x) ** 2)

# A synthesizes C with rate dependent on bound B
Rule('_R1', A() @ x >> A() @ x + C(), f_synth)

# A binds B
Rule('_R2', A(b=None) + B(a=None) | A(b=1) % B(a=1), kp, km)

# degradation of C
Rule('_R3', C() >> None, k_degrC)

Initial(A(b=MultiState(None, None, None)), Ab_b_b_0)
Initial(B(a=None), Ba_0)
Initial(C(), C_0)



from pysb import *
from pysb.macros import *
from scipy.constants import N_A

Model()

Parameter('vol', 1e-20)

NA_V = N_A*vol.value

Parameter('k1', 0.015*NA_V)
# Parameter('k1', 0.015)

Parameter('k2', 0)

Parameter('k3', 200/NA_V)
# Parameter('k3', 200)

Parameter('k4', 2*180/NA_V/NA_V)
# Parameter('k4', 2*180)

Parameter('kp4', 0.018)
Parameter('k5', 0)
Parameter('k6', 1.0)
Parameter('k7', 0.6)
Parameter('k8', 1e6) #1e12)
Parameter('k9', 1e3) #1e6)

Monomer('cyclin', ['Y', 'b'], {'Y': ['U','P']})
Monomer('cdc2', ['Y', 'b'], {'Y': ['U','P']})

# Rule 1
synthesize(cyclin(Y='U', b=None), k1)

# Rule 2
#degrade(cyclin(Y='U', b=None), k2)

# Rule 3
Rule('BindingAndPhosphoylation', cyclin(Y='U', b=None) + cdc2(Y='P', b=None) >> cyclin(Y='P', b=1) % cdc2(Y='P', b=1), k3)

# Rule 4
Rule('Activation', cyclin(Y='P', b=1) % cdc2(Y='P', b=1) >> cyclin(Y='P', b=1) % cdc2(Y='U', b=1), kp4)

# Rule 4'
Rule('Autocatalytic', cyclin(Y='P', b=1) % cdc2(Y='P', b=1) + cyclin(Y='P', b=2) % cdc2(Y='U', b=2) + cyclin(Y='P', b=2) % cdc2(Y='U', b=2) >>
cyclin(Y='P', b=1) % cdc2(Y='U', b=1) + cyclin(Y='P', b=2) % cdc2(Y='U', b=2) + cyclin(Y='P', b=2) % cdc2(Y='U', b=2), k4)

# Rule 5
#Rule('Opposed', cyclin(Y='P', b=1) % cdc2(Y='U', b=1) >> cyclin(Y='P', b=1) % cdc2(Y='P', b=1), k5)

# Rule 6
#Rule('Dissociation', cyclin(Y='P', b=1) % cdc2(Y='U', b=1) >> cyclin(Y='P', b=None) + cdc2(Y='U', b=None), k6)
Rule('Dissociation', cyclin(Y='P', b=1) % cdc2(Y='U', b=1) >> cdc2(Y='U', b=None), k6)

# Rule 7
#degrade(cyclin(Y='P', b=None), k7)

# Rules 8 and 9
equilibrate(cdc2(Y='U', b=None), cdc2(Y='P', b=None), [k8, k9])

Observable("YT", cyclin()) # Total Cyclin
Observable("CT", cdc2()) # Total CDC2
Observable("M", cyclin(Y='P', b=1) % cdc2(Y='U', b=1) ) # Active Complex

Observable("Y1", cdc2(Y='U', b=None))
Observable("Y2", cdc2(Y='P', b=None))
Observable("Y3", cdc2(Y='U', b=1) % cyclin(Y='P', b=1))
Observable("Y4", cdc2(Y='P', b=1) % cyclin(Y='P', b=1))
Observable("Y5", cyclin(Y='U', b=None))

#Observable("CYCLIN_P", cyclin(Y='P', b=None))
# [C2] in Tyson

Parameter("cdc0", 1*NA_V)
# Parameter("cdc0", 1.0)
Initial(cdc2(Y='P', b=None), cdc0)

# [Y] in Tyson
Parameter('cyc0', 0.25*NA_V)
# Parameter('cyc0', 0.25)
Initial(cyclin(Y='U', b=None), cyc0)



from pysb import *
from pysb.testing import *
from pysb.bng import *
from pysb.integrate import Solver
import numpy as np
from sympy import Piecewise


@with_model
def test_ic_expression_with_two_parameters():
    Monomer('A')
    Parameter('k1', 1)
    Parameter('k2', 2)
    Expression('e1', k1*k2)
    Rule('A_deg', A() >> None, k1)
    Initial(A(), e1)
    generate_equations(model)
    t = np.linspace(0, 1000, 100)
    sol = Solver(model, t, use_analytic_jacobian=True)
    sol.run()

@with_model
def test_ic_expression_with_one_parameter():
    Monomer('A')
    Parameter('k1', 1)
    Expression('e1', k1)
    Rule('A_deg', A() >> None, k1)
    Initial(A(), e1)
    generate_equations(model)
    t = np.linspace(0, 1000, 100)
    sol = Solver(model, t, use_analytic_jacobian=True)
    sol.run()

@with_model
def test_expressions_with_one_observable():
    Monomer('A')
    Parameter('k1', 1)
    Observable('o1', A())
    Expression('e1', o1)
    Rule('A_deg', A() >> None, k1)
    Initial(A(), k1)
    generate_equations(model)
    t = np.linspace(0, 1000, 100)
    sol = Solver(model, t, use_analytic_jacobian=True)
    sol.run()

@with_model
def test_nested_expression():
    Monomer(u'A')
    Monomer(u'B')
    Parameter(u'k1', 1)
    Observable(u'o1', B())
    Expression(u'e1', o1*k1)
    Expression(u'e2', e1+5)
    Initial(A(), Parameter(u'A_0', 1000))
    Initial(B(), Parameter(u'B_0', 1))
    Rule(u'A_deg', A() >> None, e2)
    generate_equations(model)
    t = np.linspace(0, 1000, 100)
    sol = Solver(model, t, use_analytic_jacobian=True)
    sol.run()

@with_model
def test_piecewise_expression():
    Monomer('A')
    Observable('A_total', A())
    Expression('A_deg_expr', Piecewise((0, A_total < 400.0),
                                       (0.001, A_total < 500.0),
                                       (0.01, True)))
    Initial(A(), Parameter('A_0', 1000))
    Rule('A_deg', A() >> None, A_deg_expr)
    generate_equations(model)
    t = np.linspace(0, 1000, 100)
    sol = Solver(model, t, use_analytic_jacobian=True)
    sol.run()



from pysb.simulator import KappaSimulator
from pysb.kappa import run_simulation
from pysb.examples import michment
from pysb.bng import generate_equations
import numpy as np
from nose.tools import raises


KAPPA_SEED = 123


def _compare_kappa_sims(run_simulation_output, kappasimulator_output):
    sim1 = run_simulation_output.view('<f8')[:, 1:]
    sim2 = kappasimulator_output.observables.view('f8').reshape(sim1.shape)

    assert np.allclose(sim1, sim2, atol=1e-16, rtol=1e-16)


def test_kappa_sim_michment():
    perturbation = '%init: 12 E(s[.])'

    orig_sim = run_simulation(michment.model, time=100, points=100,
                              seed=KAPPA_SEED, perturbation=perturbation)

    sim = KappaSimulator(michment.model, tspan=np.linspace(0, 100, 101))
    x = sim.run(seed=KAPPA_SEED, perturbation=perturbation)

    _compare_kappa_sims(orig_sim, x)


@raises(ValueError)
def test_kappa_sim_invalid_arg():
    sim = KappaSimulator(michment.model, tspan=range(10))
    sim.run(spam='eggs')


def test_kappa_2sims():
    sim = KappaSimulator(michment.model, tspan=np.linspace(0, 100, 101))
    res = sim.run(n_runs=2, seed=KAPPA_SEED)
    assert res.nsims == 2


def test_kappa_2initials():
    sim = KappaSimulator(michment.model, tspan=np.linspace(0, 100, 101))
    res = sim.run(initials={
        michment.model.initials[0].pattern: [10, 100],
        michment.model.initials[1].pattern: [100, 1000]
    }, seed=KAPPA_SEED)
    assert res.nsims == 2


def test_kappa_2params():
    base_param_values = np.array(
        [p.value for p in michment.model.parameters] * 2
    ).reshape(2, len(michment.model.parameters))
    sim = KappaSimulator(michment.model, tspan=np.linspace(0, 100, 101))

    res = sim.run(param_values=base_param_values, seed=KAPPA_SEED)
    assert res.nsims == 2
    assert res.dataframe.loc[0].equals(res.dataframe.loc[1])


def test_kappa_1timepoint():
    michment.model.reset_equations()
    sim = KappaSimulator(michment.model, tspan=[0, 1])
    # This set of parameter values causes Kappa to abort after 1st time point
    res = sim.run(param_values=[100, 100, 10, 10, 100, 100], seed=KAPPA_SEED)
    assert res.dataframe.shape == (1, 4), res.dataframe.shape


def test_kappa_1timepoint_with_netgen():
    generate_equations(michment.model)
    sim = KappaSimulator(michment.model, tspan=[0, 1])
    # This set of parameter values causes Kappa to abort after 1st time point
    res = sim.run(param_values=[100, 100, 10, 10, 100, 100], seed=KAPPA_SEED)
    assert res.dataframe.shape == (1, 4), res.dataframe.shape